import json
from datetime import datetime, timedelta

import urllib3  # type: ignore

from .worker import BackgroundWorker
from ._types import MYPY


if MYPY:
    from typing import Any
    from typing import Optional

    from ._types import Event


class Transport(object):
    """Baseclass for all transports.
    A transport is used to send an event to sentry.
    """

    def __init__(self, endpoint):
        # type: () -> None
        self.endpoint = endpoint

    def capture_event(
        self,
        event  # type: Event
    ):
        # type: (...) -> None
        """This gets invoked with the event dictionary when an event should
        be sent to sentry.
        """
        raise NotImplementedError()

    def flush(
        self,
        timeout,  # type: float
        callback=None,  # type: Optional[Any]
    ):
        # type: (...) -> None
        """Wait `timeout` seconds for the current events to be sent out."""
        pass

    def kill(self):
        # type: () -> None
        """Forcefully kills the transport."""
        pass

    def __del__(self):
        # type: () -> None
        try:
            self.kill()
        except Exception:
            pass


class HttpTransport(Transport):
    """The default HTTP transport."""

    def __init__(self, endpoint):
        # type: () -> None

        super().__init__(endpoint)
        self._worker = BackgroundWorker()
        self._disabled_until = None  # type: Optional[datetime]
        self._retry = urllib3.util.Retry()

        self._pool = self._make_pool()

    def _preform_event_send(
        self,
        endpoint,
        event  # type: Event
    ):
        # type: (...) -> None
        if self._disabled_until is not None:
            if datetime.utcnow() < self._disabled_until:
                return
            self._disabled_until = None

        # print('Sending event: ')
        # print(event)

        response = self._send_event_request(self.endpoint, event)

        # print('Got response with status: ' + str(response.status))
        # print(response.data)

        try:
            if response.status == 429:
                self._disabled_until = datetime.utcnow() + timedelta(
                    seconds=self._retry.get_retry_after(response) or 60
                )
                return

            elif response.status >= 300 or response.status < 200:
                print(
                    "Unexpected status code: %s (body: %s)",
                    response.status,
                    response.data,
                )
        finally:
            response.close()

        self._disabled_until = None

    def _send_event_request(
            self,
            endpoint,  # type: str
            event,  # type: Event
    ):
        encoded_data = json.dumps(event).encode('utf-8')
        response = self._pool.request(
            "POST",
            endpoint,
            body=encoded_data,
            headers={'Content-Type': 'application/json'})

        return response

    @staticmethod
    def _make_pool():
        return urllib3.PoolManager()

    def capture_event(
        self,
        event  # type: Event
    ):
        # type: (...) -> None

        def send_event_wrapper():
            # type: () -> None
            self._preform_event_send(self.endpoint, event)

        self._worker.submit(send_event_wrapper)

    def flush(
        self,
        timeout,  # type: float
        callback=None,  # type: Optional[Any]
    ):
        # type: (...) -> None
        print("Flushing HTTP transport")
        if timeout > 0:
            self._worker.flush(timeout, callback)

    def kill(self):
        # type: () -> None
        print("Killing HTTP transport")
        self._worker.kill()


def make_transport(endpoint):
    # type: () -> Transport
    return HttpTransport(endpoint)



from django.conf import settings


def is_analytics_sending_disabled():
    if 'DEBUG_DISABLE_ANALYTICS' in settings.RIGID_ANALYTICS \
            and settings.RIGID_ANALYTICS['DEBUG_DISABLE_ANALYTICS'] is not None:
        return settings.RIGID_ANALYTICS['DEBUG_DISABLE_ANALYTICS']
    else:
        return settings.DEBUG



import time
import re

from django.conf import settings
from urllib.parse import urljoin

from .tracker.django import Tracker
from .utils import is_analytics_sending_disabled


BACKEND_ENDPOINT_BASIS = 'https://rigidanalytics.com'


class Analytics(object):

    def __init__(self, get_response):
        self.get_response = get_response
        self.intercepted_data = {}
        self.init_tracker()

    def init_tracker(self):
        base_url = settings.RIGID_ANALYTICS.get(
            'BACKEND_ENDPOINT', BACKEND_ENDPOINT_BASIS)
        endpoint = urljoin(base_url, settings.RIGID_ANALYTICS['PROJECT_ID'])
        self.tracker = Tracker(endpoint)

    def __call__(self, request):
        if is_analytics_sending_disabled():
            return self.get_response(request)

        start_time = time.time_ns()

        response = self.get_response(request)

        end_time = time.time_ns()
        processing_time = end_time - start_time

        self.tracker.send_analytics_data(
            request, response, start_time, processing_time,
            self.intercepted_data
        )

        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        self.intercepted_data['view_name'] = view_func.__name__



import os
from threading import Thread, Lock
from time import sleep, time
import traceback


from ._compat import queue, check_thread_support
from ._types import MYPY


if MYPY:
    from queue import Queue
    from typing import Any
    from typing import Optional
    from typing import Callable


_TERMINATOR = object()


class BackgroundWorker(object):
    def __init__(self):
        # type: () -> None
        check_thread_support()
        self._queue = queue.Queue(-1)  # type: Queue[Any]
        self._lock = Lock()
        self._thread = None  # type: Optional[Thread]
        self._thread_for_pid = None  # type: Optional[int]

    @property
    def is_alive(self):
        # type: () -> bool
        if self._thread_for_pid != os.getpid():
            return False
        if not self._thread:
            return False
        return self._thread.is_alive()

    def _ensure_thread(self):
        # type: () -> None
        if not self.is_alive:
            self.start()

    def _timed_queue_join(self, timeout):
        # type: (float) -> bool
        deadline = time() + timeout
        queue = self._queue
        queue.all_tasks_done.acquire()  # type: ignore
        try:
            while queue.unfinished_tasks:  # type: ignore
                delay = deadline - time()
                if delay <= 0:
                    return False
                queue.all_tasks_done.wait(timeout=delay)  # type: ignore
            return True
        finally:
            queue.all_tasks_done.release()  # type: ignore

    def start(self):
        # type: () -> None
        with self._lock:
            if not self.is_alive:
                self._thread = Thread(
                    target=self._target, name="raven-sentry.BackgroundWorker"
                )
                self._thread.setDaemon(True)
                self._thread.start()
                self._thread_for_pid = os.getpid()

    def kill(self):
        # type: () -> None
        print("background worker got kill request")
        with self._lock:
            if self._thread:
                self._queue.put_nowait(_TERMINATOR)
                self._thread = None
                self._thread_for_pid = None

    def flush(self, timeout, callback=None):
        # type: (float, Optional[Any]) -> None
        print("background worker got flush request")
        with self._lock:
            if self.is_alive and timeout > 0.0:
                self._wait_flush(timeout, callback)
        print("background worker flushed")

    def _wait_flush(self, timeout, callback):
        # type: (float, Optional[Any]) -> None
        initial_timeout = min(0.1, timeout)
        if not self._timed_queue_join(initial_timeout):
            pending = self._queue.qsize()
            print("%d event(s) pending on flush", pending)
            if callback is not None:
                callback(pending, timeout)
            self._timed_queue_join(timeout - initial_timeout)

    def submit(self, callback):
        # type: (Callable[[], None]) -> None
        self._ensure_thread()
        self._queue.put_nowait(callback)

    def _target(self):
        # type: () -> None
        while True:
            callback = self._queue.get()
            try:
                if callback is _TERMINATOR:
                    break
                try:
                    callback()
                except Exception as e:
                    print("Failed processing job")
                    traceback.print_exc()
            finally:
                self._queue.task_done()
            sleep(0)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rigidanalytics_tracker",
    version="0.0.1",
    author="Shane Reustle",
    author_email="shane@reustle.co",
    description="Server-side analytics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/reustleco/rigidanalytics-tracker/",
    license="MIT",
    packages=setuptools.find_packages(),
    install_requires=[
        "urllib3",
    ]
)




try:
    from typing import TYPE_CHECKING as MYPY
except ImportError:
    MYPY = False


if MYPY:
    from types import TracebackType
    from typing import Any
    from typing import Callable
    from typing import Dict
    from typing import Optional
    from typing import Tuple
    from typing import Type

    ExcInfo = Tuple[
        Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]
    ]

    Event = Dict[str, Any]
    Hint = Dict[str, Any]

    Breadcrumb = Dict[str, Any]
    BreadcrumbHint = Dict[str, Any]

    EventProcessor = Callable[[Event, Hint], Optional[Event]]
    ErrorProcessor = Callable[[Event, ExcInfo], Optional[Event]]
    BreadcrumbProcessor = Callable[[Breadcrumb, BreadcrumbHint], Optional[Breadcrumb]]